/* --COPYRIGHT--,BSD
 * Copyright (c) 2017, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * --/COPYRIGHT--*/

#include <msp430.h>
#include "driverlib.h"

#include "timer.h"
#include "i2c.h"

#define I2CTIMEOUT	10000

#define SENSORBUCKCONV	0x48
#define SENSORBUCKIND	0x49
#define SENSORDIODE		0x4A
#define SENSORLED		0x4B

#define CONFIG_REG		0x01
#define HIGH_LIMIT_REG	0x02

static const char tmp_mapping[] = {SENSORBUCKCONV, SENSORBUCKIND, SENSORDIODE, SENSORLED};

int temp_flag=0;
int buck_conv_temp=0;
int buck_ind_temp=0;
int diode_temp=0;
int led_temp=0;
int temp_timer=0;
int eeprom_select=0;
int eeprom_select_flag=0, eeprom_change_flag=0;

#define MAX_TEMP 70 /** switch off, if any temperature is above that limit */

/**
 * Read 16 bit register of chip at given address
 * @param address 7 bit address of the chip at i2c bus
 * @param reg register to be read
 * @return data of the register
 */
static inline int read_reg(int address, int reg){
	char i2cdata[2];
	int timeout = 0;
	USCI_B_I2C_clearInterrupt(USCI_B0_BASE, USCI_B_I2C_NAK_INTERRUPT);
	USCI_B_I2C_setSlaveAddress(USCI_B0_BASE, address);
	USCI_B_I2C_setMode(USCI_B0_BASE, USCI_B_I2C_TRANSMIT_MODE);
	USCI_B_I2C_enable(USCI_B0_BASE);
	USCI_B_I2C_masterSendStart(USCI_B0_BASE);

	while(USCI_B_I2C_getInterruptStatus(USCI_B0_BASE, USCI_B_I2C_TRANSMIT_INTERRUPT) == 0)
		if(timeout++ > I2CTIMEOUT) {
			UCB0CTL1 |= UCTXSTP;
			return -1;
		}

	if(USCI_B_I2C_getInterruptStatus(USCI_B0_BASE, USCI_B_I2C_NAK_INTERRUPT)){
		UCB0CTL1 |= UCTXSTP;
		return -1;
	}

	USCI_B_I2C_slavePutData(USCI_B0_BASE, reg);
	timeout = 0;
	while(USCI_B_I2C_getInterruptStatus(USCI_B0_BASE, USCI_B_I2C_TRANSMIT_INTERRUPT) == 0)
		if(timeout++ > I2CTIMEOUT) return -1;

	USCI_B_I2C_masterReceiveMultiByteStart(USCI_B0_BASE);
	timeout = 0;
	while(USCI_B_I2C_getInterruptStatus(USCI_B0_BASE, USCI_B_I2C_RECEIVE_INTERRUPT) == 0)
		if(timeout++ > I2CTIMEOUT) return -1;

	i2cdata[0] = USCI_B_I2C_masterReceiveMultiByteFinish(USCI_B0_BASE);
	timeout = 0;
	while(USCI_B_I2C_getInterruptStatus(USCI_B0_BASE, USCI_B_I2C_RECEIVE_INTERRUPT) == 0)
		if(timeout++ > I2CTIMEOUT) return -1;

	i2cdata[1] = USCI_B_I2C_masterReceiveMultiByteNext(USCI_B0_BASE);

	return i2cdata[0]<<8 | i2cdata[1];
}

/**
 * Write 16 bit I2C register
 * @param address 7 bit address of the chip at i2c bus
 * @param reg register to be written
 * @param value data that should be written
 */
static void write_reg(int address, int reg, int value){
	int timeout = I2CTIMEOUT;
	USCI_B_I2C_clearInterrupt(USCI_B0_BASE, USCI_B_I2C_NAK_INTERRUPT);
	USCI_B_I2C_setSlaveAddress(USCI_B0_BASE, address);
	USCI_B_I2C_setMode(USCI_B0_BASE, USCI_B_I2C_TRANSMIT_MODE);
	USCI_B_I2C_enable(USCI_B0_BASE);

	USCI_B_I2C_masterSendMultiByteStartWithTimeout(USCI_B0_BASE, reg, I2CTIMEOUT);
	USCI_B_I2C_masterSendMultiByteNextWithTimeout(USCI_B0_BASE, value>>8, I2CTIMEOUT);
	USCI_B_I2C_masterSendMultiByteFinishWithTimeout(USCI_B0_BASE, value, I2CTIMEOUT);
	while((UCB0STAT & UCBBUSY) && (timeout--));
}

/**
 * read temperature from TMP116
 * @param address address of TMP116
 * @return temperature in deg C
 */
static int get_temp(int address){
	int temp = read_reg(address, 0)/128;
	if(temp > MAX_TEMP){
		operation_state = 0;
		error = fail_temp;
	}

	return temp;
}

/**
 * read EEPROM of TMP116
 * @param data pointer to store data in, 4 * 16 bit necessary
 */
void read_eeprom(int *data){
	int i;
	for(i=0;i<4;i++){
		data[i] = read_reg(tmp_mapping[eeprom_select], 5+i);
	}
}

/**
 * Write complete EEPROM at once
 * @param data data to be written, 4 * 16 bit
 */
void write_eeprom_all(int *data){
	int i;
	write_reg(tmp_mapping[eeprom_select], 4, (1<<15)); //unlock eeprom
	for(i=0;i<4;i++){
		while(read_reg(tmp_mapping[eeprom_select], 4) & (1<<14)); // wait while busy;
		write_reg(tmp_mapping[eeprom_select],5+i, data[i]);
	}
	while(read_reg(tmp_mapping[eeprom_select], 4) & (1<<14)); // wait while busy;
}

/**
 * retrieve temperature from one of the TMP116 on TIDA1081 each call
 */
void poll_temp(void){
	static int state=0;
#ifdef DEBUG
	PJOUT |= (1<<3);
#endif
	switch(state++ & 0x3){
	case 0:
		buck_conv_temp = get_temp(SENSORBUCKCONV);
		break;
	case 1:
		buck_ind_temp = get_temp(SENSORBUCKIND);
		break;
	case 2:
		diode_temp = get_temp(SENSORDIODE);
		break;
	case 3:
		led_temp = get_temp(SENSORLED);
		break;
	}
	temp_flag = 1;
#ifdef DEBUG
	PJOUT &= ~(1<<3);
#endif
}

/**
 * Resets temperature error flag
 */
void reset_temp_error(void){
	read_reg(SENSORBUCKCONV, CONFIG_REG);
	read_reg(SENSORBUCKIND, CONFIG_REG);
	read_reg(SENSORDIODE, CONFIG_REG);
	read_reg(SENSORLED, CONFIG_REG);
}

/**
 * Initialize I2C and set limits for alert output
 */
void i2c_init(void){
	USCI_B_I2C_initMasterParam param = {0};
	param.selectClockSource = USCI_B_I2C_CLOCKSOURCE_SMCLK;
	param.i2cClk = UCS_getSMCLK();
	param.dataRate = USCI_B_I2C_SET_DATA_RATE_400KBPS;
	USCI_B_I2C_initMaster(USCI_B0_BASE, &param);

	// configure alert for TMP116
	write_reg(SENSORBUCKCONV, HIGH_LIMIT_REG, MAX_TEMP*128);
	write_reg(SENSORBUCKIND, HIGH_LIMIT_REG, MAX_TEMP*128);
	write_reg(SENSORDIODE, HIGH_LIMIT_REG, MAX_TEMP*128);
	write_reg(SENSORLED, HIGH_LIMIT_REG, MAX_TEMP*128);
}
