/* --COPYRIGHT--,BSD
 * Copyright (c) 2017, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * --/COPYRIGHT--*/

#include <msp430.h>
#include "driverlib.h"
#include "timer.h"
#include "vt100.h"
#include "i2c.h"

static const uint8_t port1_mapping[] = {
	PM_ANALOG,
	PM_ANALOG,
	PM_ANALOG,
	PM_ANALOG,
	PM_UCB0SDA,
	PM_UCB0SCL,
	PM_TD1_1,
	PM_TD1_2,
};

static const uint8_t port2_mapping[] = {
	PM_NONE,
	PM_NONE,
	PM_UCA0RXD, // 2.2
	PM_UCA0TXD, // 2.3
	PM_NONE,
	PM_TD1_2, //PM_NONE, //Buck COff
#ifdef TRIGGER_GPIO
	PM_NONE,
#else
	PM_TD1_2,
#endif
	PM_NONE,
};

static const uint8_t port3_mapping[] = {
	PM_NONE,
	PM_NONE,
	PM_NONE,
	PM_NONE,// soft spi sclk
	PM_NONE,// soft spi dout
	PM_NONE,
	PM_NONE,
	PM_NONE,
};

/**
 * Initialize Port mapping and default pin states
 */
void pinmux_init(void){
	PMAP_initPortsParam initPortsParam = {0};

	initPortsParam.portMapping = port3_mapping;
	initPortsParam.PxMAPy = (uint8_t  *)&P1MAP23;
	initPortsParam.numberOfPorts = 1;
	initPortsParam.portMapReconfigure = PMAP_ENABLE_RECONFIGURATION;
	PMAP_initPorts(PMAP_CTRL_BASE, &initPortsParam);

	initPortsParam.portMapping = port2_mapping;
	initPortsParam.PxMAPy = (uint8_t  *)&P2MAP01;
	initPortsParam.numberOfPorts = 1;
	initPortsParam.portMapReconfigure = PMAP_ENABLE_RECONFIGURATION;
	PMAP_initPorts(PMAP_CTRL_BASE, &initPortsParam);

	initPortsParam.portMapping = port1_mapping;
	initPortsParam.PxMAPy = (uint8_t  *)&P1MAP01;
	initPortsParam.numberOfPorts = 1;
	initPortsParam.portMapReconfigure = PMAP_DISABLE_RECONFIGURATION;
	PMAP_initPorts(PMAP_CTRL_BASE, &initPortsParam);

#ifdef DEBUG
	//bring out MCLK for testing LPM0
	//some more debug pins accessible via JTAG header
	PJDIR |= (1<<1)|(1<<0)|(1<<2)|(1<<3);
	PJSEL |= (1<<1);
#endif

	// Trigger in
	P2IE |= (1<<7);
	P2IES |= (1<<7); // Trigger on falling edge
	P2IFG &= ~(1<<7);

	// efuse ok IRQ
	P2IE |= (1<<4);
	P2IES |= (1<<4);
	P2IFG &= ~(1<<4);

	// buck_vout_ov_reset
	P2DIR |= (1<<0);
	P2OUT &= ~(1<<0);

	// buck_dischg_coff
//	P2DIR |= (1<<5);
//	P2OUT &= ~(1<<5);

	// buck_dischg_coff -> same signal as the short FET
	P2SEL |= (1<<5);
	P2DIR |= (1<<5);

	// UCA0TXD
	P2DIR |= (1<<3);
	P2SEL |= (1<<3);

	// UCA0RXD
	P2DIR &= ~(1<<2);
	P2SEL |= (1<<2);

	P1SEL |= 0xf; // select P1.0 .. P1.3 as analog input

	P1SEL |= (1<<4)|(1<<5); // SCL and SDA as peripheral
	P1OUT |= (1<<4)|(1<<5); // SCL and SDA pull up
	P1REN |= (1<<4)|(1<<5); // SCL and SDA pull up

	//soft spi config
	// sclk and din
	P3DIR |= (1<<3)|(1<<4);
	P3OUT &= ~((1<<3)|(1<<4));

	//cs efuse and buck
	PJDIR |= (1<<5)|(1<<6);
	PJOUT |= (1<<5)|(1<<6);

	// BUCK_OFF & SHORT_LED
	P1DIR |= (1<<6)|(1<<7);
	P1OUT &= ~(1<<7);
	P1OUT |= (1<<6);

	// trigger out
	P2DIR |= (1<<6);
	P2OUT &= ~(1<<6);

	// 5V load switch
	P3OUT &= ~(1<<7);
	P3DIR |= (1<<7);

	 // buck_vout_ov_reset
	P2OUT &= ~(1<<0);
	P2DIR |= (1<<0);
	//P2OUT |= (1<<0);

	// turn eFuse on
	P3OUT |= (1<<0);
	P3DIR |= (1<<0);
}

/**
 * Set the load switch to a specific state
 * @param state set to on=1 or off=0
 */
void load_sw(int state){
	if(state)
		P3OUT |= (1<<7);
	else
		P3OUT &= ~(1<<7);
}

/**
 * Port 2 ISR - P2.7 is used as trigger in
 */
#pragma vector=PORT2_VECTOR
__interrupt void port2_isr(void)
{
	if(P2IFG & (1<<7)){ // trigger in
		if(trigger_src){
			if(timer_d_trigger()) __bic_SR_register_on_exit(LPM0_bits); // give the main loop some time
			if(TA0R > 1024){ // check timer 0 if it is time to trigger the main loop
				TA0CTL |= (1<<2);
				if(monitoring_state){
					temp_timer = 2501;
					__bic_SR_register_on_exit(LPM0_bits); // give the main loop some time
				}
			}
		}
		P2IFG &= ~(1<<7);
	}else if(P2IFG & (1<<4)){ // e fuse error
		operation_state = 0;
		monitoring_state = 1;
		error = fail_efuse;
		__bic_SR_register_on_exit(LPM0_bits);
		P2IFG &= ~(1<<4);
	}
}
