/* --COPYRIGHT--,BSD
 * Copyright (c) 2017, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * --/COPYRIGHT--*/

#include <msp430.h>
#include <stdbool.h>
#include "../main.h"
#include "timer.h"
#include "adc.h"
#include "i2c.h"
#include "dac.h"
#include "uart.h"
#include "vt100.h"

unsigned int framerate=1000;
unsigned char framerate_flag=0;
unsigned int flash_count=0;
unsigned volatile int flash_counter=0;
unsigned char flash_count_flag=0;
unsigned int ontime=1000;
unsigned char ontime_flag=0;
unsigned int timebaserange=0;
unsigned char operation_state=0; // 0 = off, 1 = on
unsigned int running=0;
unsigned char trigger_src=0; // 0 = int, 1 = ext
enum error_enum error=fail_none;
const char error_text[][6] = {"None ", "Temp ", "Boost", "eFuse"};
unsigned int flash_bracketing=0; // 0 = off
unsigned char flash_bracketing_flag=0;
unsigned char current_bracketing_flag=0;
int current_bracketing=0;
unsigned int flash_bracketing_counter=0;
const char flash_bracketing_text[][5] = {"Off ", "0.25", "0.33", "0.5 ", "1.0 ", "2.0 "};
static const float flash_bracket_time[][5] = {
		{0.707, 0.841, 1, 1.189, 1.414},
		{0.630, 0.794, 1, 1.260, 1.587},
		{0.500, 0.707, 1, 1.414, 2.000},
		{0.250, 0.500, 1, 2.000, 4.000},
		{0.063, 0.250, 1, 4.000, 16.00}
};
volatile int lf_counter=0;

#define TIMERD_IRQ_PRE	10 	/** offset for timer d irq */
#define CURRENT_OFFSET	200 /** mA for timing lookup table */
#define CURRENT_STEP	50  /** mA for timing lookup table */

#define CALTDH0CTL1_64		*((unsigned int *)0x1A30)
#define CALTDH0CTL1_128		*((unsigned int *)0x1A32)
#define CALTDH0CTL1_200		*((unsigned int *)0x1A34)
#define CALTDH0CTL1_256		*((unsigned int *)0x1A36)

#define BUCK_OFF	(1<<6)
#define LED_SHORT	(1<<7)
#define TRIGGER_OUT (1<<6)
#define COFF		(1<<5)

static int timer_ctl0;

/** precharge time for buck regulator, 200mA to 2.4A in 50mA steps, scaled for 200MHz timer d */
static const int buck_charge[45] = {198, 232, 265, 298, 331, 364, 397, 430, 463, 496, 529, 562, 595, 628,
		662, 695, 728, 761, 794, 827, 860, 893, 926, 959, 992, 1025, 1058, 1091, 1125, 1158, 1191, 1224,
		1257, 1290, 1323, 1356, 1389, 1422, 1455, 1488, 1521, 1555, 1588, 1588, 1588};

static const int ontime_correction[45] = {17, 10, 7, 5, 4, 2, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0};

static const int trigger_delay=8100/5; //8400ns extra delay until flash

/**
 * Configure Timer A0 to interrupt every 100 us for generating internal timing
 */
void timer_init(void){
	TA0CCTL0 = CCIE;                          // CCR0 interrupt enabled
	TA0CCR0 = 2500/4;
	TA0CTL = TASSEL_2 + MC_1 + TACLR + ID_2;         // SMCLK, upmode, clear TAR, divide by 4
}

/**
 * Configure timer A0 to interrupt every 5s, for generating a timeout in external trigger mode
 */
void timer_init_ext(void){
	TA0CCTL0 = CCIE;                          // CCR0 interrupt enabled
	TA0CCR0 = 4096*5;//every 5 seconds
	TA0CTL = TASSEL_1 + MC_1 + TACLR + ID_3;         // ACLK, upmode, clear TAR, divide by 8
}

/**
 * Configure Timer D to generate LED flash
 * @param on_time time LED should be on
 * @param current configured LED current, necessary for buck on before led on timing
 * @param range time range that is currently selected
 */
void timer_d_configure(int on_time, unsigned int current, int range){
	unsigned int current_step = (current-CURRENT_OFFSET)/CURRENT_STEP;

	switch(range){
		case 0:	// 100 ns - 10 us
			// tdclk, div by 1, 200MHz, 5ns resolution
			TD1CCTL0 = CCIE;

			TD1CTL1 |= TDCLKM0; // high res internal clock
			TD1CCTL1 |= OUTMOD_7;//dischg_l = 1

			TD1CCTL2 = OUTMOD_0 + OUT; // short_led set short fet on
			TD1CCTL2 |= OUTMOD_5; // short_led reset short fet

			TD1CCR0 = trigger_delay + (on_time)/4.6 + ontime_correction[current_step] + TIMERD_IRQ_PRE; // the timer is actually a bit fast than 200 MHz...
			TD1CCR1 = trigger_delay - buck_charge[current_step];
			TD1CCR2 = trigger_delay;

			TD1HCTL1 = CALTDH0CTL1_200;				// Read the 200MHz TimerD TLV Data
			TD1CTL1 |= TDCLKM_1;					// Select Hi-res local clock
			TD1HCTL0 = TDHEN + TDHM_0 +TDHRON;		// CALEN=0 => free running mode; enable Hi-res mode; THDMx = 00 => 8x
			TD1CTL0 = TDCLR; // cont mode
			timer_ctl0 = TD1CTL0|TDCLR|MC_2;
		break;
		case 1: // 10 us - 10 ms
			// smclk, div by 4, 25/4 MHz, 160ns resolution
			TD1CCTL0 = CCIE;
			TD1HCTL0 = 0;
			TD1HCTL1 = 0;
			TD1CTL1 &= ~(TDCLKM0 | TDCLKM1);
			TD1CCTL1 |= OUTMOD_7;//dischg_l = 1
			TD1CCTL2 = OUTMOD_0 + OUT; // short_led set short fet on
			TD1CCTL2 |= OUTMOD_5; // short_led reset short fet

			TD1CCR0 = trigger_delay/32 + (on_time*6.25) + TIMERD_IRQ_PRE/32;
			TD1CCR1 = trigger_delay/32 - buck_charge[current_step]/32;
			TD1CCR2 = trigger_delay/32 - 2;

			TD1CTL0 =  TDCLR | TDSSEL__SMCLK | TDHD1; // count up, div by 4, smclk
			timer_ctl0 = TD1CTL0|TDCLR|MC_2;
		break;
		case 2: // 10 ms - 10 s
			// configure to run at 1ms, handle timings in irq
			TD1CCTL0 = CCIE;
			TD1HCTL0 = 0;
			TD1HCTL1 = 0;
			TD1CTL1 &= ~(TDCLKM0 | TDCLKM1);
			TD1CCTL1 = OUTMOD_0 + OUT;
			TD1CCTL1 |= OUTMOD_5;
			TD1CCTL2 = OUTMOD_0 + OUT; // short_led set short fet on
			TD1CCTL2 |= OUTMOD_5; // short_led reset short fet

			TD1CCR0 = trigger_delay/32 + (1000*6.25) + TIMERD_IRQ_PRE/32;
			TD1CCR1 = trigger_delay/32 - buck_charge[current_step]/32;
			TD1CCR2 = trigger_delay/32 - 2;

			TD1CTL0 =  TDCLR | TDSSEL__SMCLK | TDHD1; // count up, div by 4, smclk
			timer_ctl0 = TD1CTL0|TDCLR|MC_1;
		break;
	}
}

/**
 * Trigger Timer D to flash LED once, hand over IOs to timer and assert trigger out signal
 */
inline int timer_d_trigger(void){
	if(flash_bracketing){
		long new_time = ontime*flash_bracket_time[flash_bracketing-1][flash_bracketing_counter++];
		if(new_time <= 10000)
			timer_d_configure(new_time, led_current, timebaserange);
		else
			timer_d_configure(new_time/1000, led_current, timebaserange+1);
		if(flash_bracketing_counter >= 5) {
			operation_state = 0; // set to off after sequence
		}
	}

	TD1CCTL2 = OUTMOD_0 + OUT; // short_led set short fet on
	TD1CCTL2 |= OUTMOD_5; // short_led reset short fet
	P1SEL |= BUCK_OFF|LED_SHORT;
	P2SEL |= COFF;
	TD1CTL0 = timer_ctl0; // start timer

	if(running==0) {
		TD1CTL0 = 0;
		P1OUT |= BUCK_OFF;
		P1OUT &= ~LED_SHORT;
		P1SEL &= ~(BUCK_OFF|LED_SHORT);
		return 1;
	}

	if(!(P3IN & (1<<2))) { // alert pin from TMP116
		TD1CTL0 = 0;
		P1OUT |= BUCK_OFF;
		P1OUT &= ~LED_SHORT;
		P1SEL &= ~(BUCK_OFF|LED_SHORT);
		operation_state = 0;
		monitoring_state = 1;
		error = fail_temp;
		return 1;
	}

	if(flash_count && running){
		if(flash_counter++ >= flash_count) {
			operation_state = 0;
			TD1CTL0 = 0;
			P1OUT |= BUCK_OFF;
			P1OUT &= ~LED_SHORT;
			P1SEL &= ~(BUCK_OFF|LED_SHORT);
			return 1;
		}
	}

	if(framerate == 0){//check pin -> low: dc on, irq rising -> high: dc off, irq falling
		if(P2IN & (1<<7)){
			dc_mode(0);
			P2IES |= (1<<7); // Trigger on falling edge
			P2IFG &= ~(1<<7);
		}else{
			dc_mode(1);
			P2IES &= ~(1<<7); // Trigger on rising edge
			P2IFG &= ~(1<<7);
		}

	}

	lf_counter=0;
	P2OUT |= TRIGGER_OUT;
	return 0;
}

/**
 * Reconfigures timings of timer D to generate LED flash
 * @param on_time time LED should be on
 * @param current configured LED current, necessary for buck on before led on timing
 * @param range time range that is currently selected
 */
inline void timer_d_reconfigure(int on_time, unsigned int current, int range){
	unsigned int current_step = (current-CURRENT_OFFSET)/CURRENT_STEP;
	switch(range){
		case 0:	// 100 ns - 10 us
			TD1CCR0 = trigger_delay + (on_time)/5 + ontime_correction[current_step] + TIMERD_IRQ_PRE; // the timer is actually a bit fast than 200 MHz...
			TD1CCR1 = trigger_delay - buck_charge[current_step];
			TD1CCR2 = trigger_delay;
		break;
		case 1: // 10 us - 10 ms
			TD1CCR0 = trigger_delay/32 + (on_time*6) + TIMERD_IRQ_PRE/32;
			TD1CCR1 = trigger_delay/32 - buck_charge[current_step]/32;
			TD1CCR2 = trigger_delay/32;
		break;
		case 2: // 10 ms - 10 s
			TD1CCR0 = 2 + on_time*4;
			TD1CCR1 = 1;
			TD1CCR2 = 2;
		break;
	}
}

/**
 * Timer D ISR, disabling timer and setting IOs back to GPIO, in 10ms - 10s range manual timing
 */
#pragma vector=TIMER1_D0_VECTOR
__interrupt void TD1_isr(void)
{
	static int led_current_temp;
	if(timebaserange == 2 && running ){
		if(lf_counter++ == 0){
			//manual control, buck on, short off
			P1OUT &= ~BUCK_OFF;
			P1OUT &= ~LED_SHORT;
			P1SEL &= ~(BUCK_OFF|LED_SHORT);
		}else if(lf_counter == ontime){ // stop everything
			P1OUT |= BUCK_OFF;
			P1OUT &= ~LED_SHORT;
			P2OUT &= ~TRIGGER_OUT;
			TD1CTL0 = TDCLR | MC_0;
		}
	}else{
		P1OUT |= BUCK_OFF;
		P1OUT &= ~LED_SHORT;
		P1SEL &= ~(BUCK_OFF|LED_SHORT);
		P2OUT &= ~TRIGGER_OUT;
		TD1CTL0 = TDCLR | MC_0;
		if(flash_counter>0 && current_bracketing!=0){
			if(flash_counter == 1) led_current_temp = led_current + current_bracketing;
			else led_current_temp += current_bracketing;
			if(flash_counter >= flash_count) {
				dac73_set(led_current);
				timer_d_reconfigure(ontime, led_current, timebaserange);
			}else{
				if(led_current_temp < 200) led_current_temp = 200;
				if(led_current_temp > 2400) led_current_temp = 2400;
				dac73_set(led_current_temp);
				timer_d_reconfigure(ontime, led_current_temp, timebaserange);
			}

		}
	}
	//__bic_SR_register_on_exit(LPM0_bits);   // Exit active CPU
}

/**
 * Timer A0 ISR, main timer at 10khz for triggering cyclic flashing, adc, temperature measurement,
 */
#pragma vector=TIMER0_A0_VECTOR
__interrupt void ta0_isr(void)
{
#ifdef DEBUG
	PJOUT |= (1);
#endif
	if(trigger_src == 1) {
		temp_timer = 2501;
		__bic_SR_register_on_exit(LPM0_bits); // give the main loop some time
	}
	else {
		static int trigger_timer=0;
		static int main_loop=0;
		if((framerate != 0) && (trigger_src == 0) && operation_state){
			if(trigger_timer++ >= (framerate-1)){
				timer_d_trigger();
				trigger_timer = 0;
			}
		}
		temp_timer++;
		if(main_loop++ > 10000/4)  {
			__bic_SR_register_on_exit(LPM0_bits); // let main loop run ever 250ms for checking new ADC values
			main_loop = 0;
		}
		// Synchronize UART transmission to timer IRQ, so UART activity does not interfere timer IRQ
		if(tx_flag){
			if(*tx_ptr != 0)
				UCA0TXBUF = *tx_ptr++; // transmit next char from ADC buffer and go back to sleep
			else{
				tx_flag = 0;
				__bic_SR_register_on_exit(LPM0_bits); // when finished sending, let the CPU active to process the main loop further
			}
		}
	}
#ifdef DEBUG
	PJOUT &= ~(1);
#endif
}

/**
 * configure LED DC mode
 * @param on state of LEDs
 */
void dc_mode(bool on){
	if(on){
		P1SEL &= ~(BUCK_OFF|LED_SHORT);
		P2SEL &= ~(COFF);
		P2OUT &= ~(COFF);
		P1OUT &= ~LED_SHORT;
		P1OUT &= ~BUCK_OFF;
		TD1CTL0 = 0;
	}else{
		P1SEL &= ~(BUCK_OFF|LED_SHORT);
		P2SEL &= ~(COFF);
		P2OUT &= ~(COFF);
		P1OUT &= ~LED_SHORT;
		P1OUT |= BUCK_OFF;
	}
}
