/* --COPYRIGHT--,BSD
 * Copyright (c) 2017, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * --/COPYRIGHT--*/

#include <msp430.h>
#include "driverlib.h"

#include "vt100.h"
#include "timer.h"

#include <stdio.h>
#include <stdarg.h>

volatile char *tx_ptr=0; 	// pointer to UART data, used in timer ISR for synchronized transmission
volatile int tx_flag;		// indicated an active UART transmission
volatile char input_buffer[8];
volatile int input_buffer_pos=0;

/**
 * Init USCIA0 as UART with 115200 Baud
 */
void uart_init(void){
	//Baudrate = 115200, clock freq = 25MHz
	//UCBRx = 13, UCBRFx = 9, UCBRSx = 0, UCOS16 = 1
	USCI_A_UART_initParam param = { 0 };
	param.selectClockSource = USCI_A_UART_CLOCKSOURCE_SMCLK;
	param.clockPrescalar = 13;
	param.firstModReg = 9;
	param.secondModReg = 0;
	param.parity = USCI_A_UART_NO_PARITY;
	param.msborLsbFirst = USCI_A_UART_LSB_FIRST;
	param.numberofStopBits = USCI_A_UART_ONE_STOP_BIT;
	param.uartMode = USCI_A_UART_MODE;
	param.overSampling = USCI_A_UART_OVERSAMPLING_BAUDRATE_GENERATION;

	if (STATUS_FAIL == USCI_A_UART_init(USCI_A0_BASE, &param)) {
		return;
	}

	//Enable UART module for operation
	USCI_A_UART_enable(USCI_A0_BASE);

	//Enable Receive Interrupt
	USCI_A_UART_clearInterrupt(USCI_A0_BASE, USCI_A_UART_RECEIVE_INTERRUPT);
	USCI_A_UART_enableInterrupt(USCI_A0_BASE, USCI_A_UART_RECEIVE_INTERRUPT);

	__enable_interrupt();
}

/**
 * Write one char to UART
 * @param c char to write
 */
void uart_putc(char c){
	USCI_A_UART_transmitData(USCI_A0_BASE, c);
}

/**
 * Write string to UART
 * @param s string to write
 */
void uart_puts(char *s){
	while (tx_flag) __bis_SR_register(LPM0_bits + GIE);
	while(*s != 0)
		uart_putc(*s++);
}

/**
 * Write string to UART synchronized to timer ISR
 * @param s string to write
 */
void uart_puts_sync(char *s){
	while(!USCI_A_UART_getInterruptStatus(USCI_A0_BASE, UCTXIFG));
	tx_flag = 1;
	tx_ptr = s;
	__bis_SR_register(LPM0_bits + GIE);
	while(tx_flag);
}

/**
 * Printf to UART
 * @param format format to print
 */
void uart_printf(const char * format, ...){
	char buffer[64];
	va_list args;
	va_start(args, format);
	vsprintf(buffer,format, args);
	if(trigger_src) uart_puts(buffer);  // external trigger -> output immediately
	else uart_puts_sync(buffer);		// internal trigger -> data output synchronized to timer IRQ
	va_end(args);
}

/**
 * USCIA0 ISR receive char and handle
 */
#pragma vector=USCI_A0_VECTOR
__interrupt void uscia_isr(void){
	if(UCA0IFG & UCRXIFG){
		UCA0IFG &= ~UCRXIFG;
		input_buffer[input_buffer_pos++%sizeof(input_buffer)] = UCA0RXBUF;
	}
	__bic_SR_register_on_exit(LPM0_bits);   // keep the CPU on to run through the main loop once to handle new parameters
}
