/* --COPYRIGHT--,BSD
 * Copyright (c) 2017, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * --/COPYRIGHT--*/

#include <msp430.h>
#include "uart.h"
#include "timer.h"
#include "vt100.h"
#include "i2c.h"
#include "dac.h"

#include "../sanity.h"

#include <stdio.h>
#include <stdarg.h>
#include <string.h>

#define ESC 		"\033"
#define CSI 		ESC"["
#define UNDERLINE	CSI"4m"
#define RESETFORMAT	CSI"0m"
#define CURSOROFF	CSI"?25l"
#define CURSORON	CSI"?25h"

#define UNDERLINECHAR(x)	UNDERLINE x RESETFORMAT

#define TABSTART	30
#define VALUESTART	15

static int edit_flag=0;
static volatile int write_flag=0;
int monitoring_state=1;
int menu_init_state=0;
static char ontime_value[10] = "";
static char vcc_value[3] = "ok ";
static char vin_value[10];
static char iin_value[11];
static char vboost_value[10];
static char vled_value[10];
static char led_temp_value[6];
static char diodes_temp_value[6];
static char buck_temp_value[6];
static char inductor_temp_value[6];
static char boost_ovp_value[6];
static char buck_ovp_value[6];
static char efuse_value[5];
static char loadsw_value[4] = "on ";
/*static char efuse_dac_value[7];
static char boost_dac_value[7];
static char buck_dac_value[7];*/
static char framerate_value[11] = "";
static char state_value[4] = "";
static char eeprom_value[2] = "0";
static char eeprom_data_value[20];
static char led_current_value[9]="";
static char trigger_value[4];
static char number_value[5];
static char flash_bracketing_value[10];
static char monitoring_value[4]="On ";
static char overtemp_value[5];
static char fail_value[10]="None";
static char current_bracketing_value[10]="";

const struct menu_struct menu[]={
	{"Control:", 0, 1, ""},										{"Temp:"		, TABSTART, 1, ""},
	{" "UNDERLINECHAR("O")"n Time [s]", 0, 2, ontime_value},	{" LED"		, TABSTART, 2, led_temp_value},
	{" "UNDERLINECHAR("F")"ramerate", 0, 3, framerate_value},	{" Diodes"		, TABSTART, 3, diodes_temp_value},
	{" "UNDERLINECHAR("T")"rigger", 0, 4, trigger_value},		{" TPS92515"	, TABSTART, 4, buck_temp_value},
	{" "UNDERLINECHAR("N"), 0, 5, number_value},				{" Inductor"	, TABSTART, 5, inductor_temp_value},
	{" "UNDERLINECHAR("I")" LED", 0, 6, led_current_value},		{"Status:"		, TABSTART, 7, ""},
	{"", 0, 8, ""},												{" V in"		, TABSTART, 8, vin_value},
	{"", 0, 9, ""},												{" I in"		, TABSTART, 9, iin_value},
	{"", 0, 10, ""},											{" V boost"		, TABSTART, 10, vboost_value},
	{"", 0, 11, ""},											{" V LED"		, TABSTART, 11, vled_value},
																{" 3.3 V"		, TABSTART, 12, vcc_value},
	{UNDERLINECHAR("R")"eset OV", 0, 13, ""},					{" Buck OVP"	, TABSTART, 13, buck_ovp_value},
	{UNDERLINECHAR("S")"tate", 0, 14, state_value},				{" Boost OVP"	, TABSTART, 14, boost_ovp_value},
																{" eFuse"		, TABSTART, 15, efuse_value},
	{UNDERLINECHAR("E")"EPROM", 0, 17, eeprom_value},			{" Load SW"		, TABSTART, 16, loadsw_value},
	{" "UNDERLINECHAR("D")"ata"		, 0, 18, eeprom_data_value},
	{" Time "UNDERLINECHAR("B")"rk"		, 0, 8, flash_bracketing_value},
	{" "UNDERLINECHAR("M")"onitoring"		, 0, 10, monitoring_value}, {" Over Temp"		, TABSTART, 17, overtemp_value},
																{" Fail"		, TABSTART, 19, fail_value},
	{" "UNDERLINECHAR("C")"urrent Brk"		, 0, 9, current_bracketing_value},
};

/**
 * Initialize terminal as vt100 terminal
 */
void vt100_init(void){
	uart_printf(CSI"37;40m" CSI"2J" CSI"H" CSI"6n" CSI"2J");
	uart_printf(CURSOROFF);	//cursor off
}

/**
 * Set cursor to position
 * @param x x coordinate
 * @param y y coordinate
 */
void vt100_setcur(int x, int y){
	uart_printf(CSI"%d;%dH", y, x);
}

/**
 * Printf over UART to specific position
 * @param x position
 * @param y position
 * @param format format to print
 */
void vt100_printf(int x, int y, const char * format, ...){
	char buffer[64];
	va_list args;
	va_start(args, format);
	vsprintf(buffer,format, args);
	vt100_setcur(x, y);
	if(trigger_src) uart_puts(buffer);  // external trigger -> output immediately
	else uart_puts_sync(buffer);		// internal trigger -> data output synchronized to timer IRQ
	va_end(args);
}

/**
 * Initialize menu on serial line by writing every field
 */
void menu_init(void){
	int i;
	vt100_init();
	for(i=0; i<(sizeof(menu)/sizeof(menu[0])); i++){
		menu_write_field((enum field_enum)i);
	}
}

/**
 * Edit one field by setting the cursor to the position
 * @param field field to edit
 */
void menu_edit(enum field_enum field){
	edit_flag = 1;
	vt100_setcur(menu[field].startx+VALUESTART, menu[field].starty);
	uart_printf(CURSORON);
}

/**
 * Write the value of a field to console
 * @param field field to update
 */
void menu_write_field(enum field_enum field){
	if(menu_init_state < 2) return;
	if(!edit_flag){
		write_flag = 1;
		vt100_printf(menu[field].startx, menu[field].starty, menu[field].text);
		vt100_printf(menu[field].startx+VALUESTART, menu[field].starty, menu[field].value);
		write_flag = 0;
	}
}

/**
 * Updates the value of a field
 * @param field field to update
 */
void menu_update_field(enum field_enum field){
	if(menu_init_state < 2) return;
	if(!edit_flag){
		write_flag = 1;
		vt100_printf(menu[field].startx+VALUESTART, menu[field].starty, menu[field].value);
		write_flag = 0;
	}
}

/**
 * Handle each entered char
 * @param c char received
 */
void menu_input(char c){
	static char input[32];
	static int pos=0;
	int value1=0, value2=0;
	int digit=0;
	char unit;
	if(write_flag & trigger_src==0) return;
	input[pos++%sizeof(input)] = c;

	switch(input[0]){
		case 'o':
		case 'O':
			if(pos == 1)
				menu_edit(ontime_field);
			else if(c == '\r' || c == '\n'){
				input[pos-1] = 0;

				if(sscanf(&input[1], "%d.%d%c", &value1, &value2, &unit) == 3){
					digit = 1;
				}else{
					sscanf(&input[1], "%d%c", &value1, &unit);
				}

				switch(unit){
				case 'u':
					if(value1 < 10) timebaserange = 0;
					else timebaserange = 1;
					break;
				case 'm':
					if(value1 < 10) timebaserange = 1;
					else timebaserange = 2;
					break;
				case 'n':
					timebaserange = 0;
					break;
				default:
					if(value1 < 10) timebaserange = 2;
					else timebaserange = 3;
					break;
				}

				if(value1 < 10 && unit != 'n') {
					ontime = value1*1000;
					if(digit) ontime += value2*100;
				}
				else ontime = value1;
				ontime_flag = 1;
			}else if(pos < 8)
				uart_putc(c);
			break;

		case 'f':
		case 'F':
			if(pos == 1)
				menu_edit(framerate_field);
			else if(c == '\r' || c == '\n'){
				input[pos-1] = 0;
				digit = sscanf(&input[1], "%d.%d", &value1, &value2);
				if(digit==1){
					if(value1 != 0)
						framerate = 10000/value1;
					else
						framerate = 0;
				}else{
					value1 = value1*10 + value2;
					if(value1 != 0)
						framerate = 100000.0/value1;
					else
						framerate = 0;
				}

				framerate_flag = 1;
			}else if(pos < 7)
				uart_putc(c);
			break;

		case 's':
		case 'S':
			if(pos == 1){
				pos = 0;
				operation_state ^= 1;
			}
			break;

		case 'E':
		case 'e':
			if(pos == 1)
				menu_edit(eeprom_field);
			else if(c == '\r' || c == '\n'){
				input[pos-1] = 0;
				sscanf(&input[1], "%d", &eeprom_select);
				eeprom_select_flag = 1;
			}else if(pos < 3)
				uart_putc(c);
			break;

		case 'D':
		case 'd':
			if(pos == 1)
				menu_edit(eeprom_data_field);
			else if(c == '\r' || c == '\n'){
				input[pos-1] = 0;
				strcpy(menu[eeprom_data_field].value, &input[1]);
				eeprom_change_flag = 1;
			}else if(pos < 21)
				uart_putc(c);
			break;

		case 'i':
		case 'I':
			if(pos == 1)
				menu_edit(led_current_field);
			else if(c == '\r' || c == '\n'){
				input[pos-1] = 0;
				sscanf(&input[1], "%d", &led_current);
				led_current_flag = 1;
			}else if(pos < 7)
				uart_putc(c);
			break;

		case 't':
		case 'T':
			if(pos == 1){
				pos = 0;
				if(!operation_state) // only allowed if not turned on
					trigger_src ^= 1;
			}
			break;

		case 'r':
		case 'R':
			if(pos == 1){
				P2OUT &= ~(1<<0);
				pos = 0;
				P2OUT |= (1<<0);
			}
			break;

		case 'n':
		case 'N':
			if(pos == 1)
				menu_edit(nflashes_field);
			else if(c == '\r' || c == '\n'){
				input[pos-1] = 0;
				sscanf(&input[1], "%d", &flash_count);
				flash_count_flag = 1;
			}else if(pos < 7)
				uart_putc(c);
			break;

		case 'b':
		case 'B':
			if(pos == 1){
				pos = 0;
				flash_bracketing = (flash_bracketing+1) % 6;
				flash_bracketing_flag = 1;
			}
			break;

		case 'v':
			if(pos == 1){
				pos = 0;
				menu_init();
			}
			break;

		case 'M':
		case 'm':
			if(pos == 1){
				pos = 0;
				monitoring_state ^= 1;
			}
			break;

		case 'C':
		case'c':
			if(pos == 1)
				menu_edit(current_bracketing_field);
			else if(c == '\r' || c == '\n'){
				input[pos-1] = 0;
				sscanf(&input[1], "%d", &current_bracketing);
				current_bracketing_flag = 1;
			}else if(pos < 7)
				uart_putc(c);
			break;
		default:
			pos = 0;
			break;
	}

	if(c == '\r' || c== '\n'){
		uart_printf(CURSOROFF);
		edit_flag = 0;
		pos = 0;
	}
	if(c == '['){
		input_buffer_pos = 0;
	}
}

