/* --COPYRIGHT--,BSD
 * Copyright (c) 2017, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * --/COPYRIGHT--*/

#include "drv/timer.h"
#include "drv/dac.h"

/**
 * @brief Structure for defining a range
 */
struct ontime_range_struct{
	unsigned int min;		/** Minimal value in this range */
	unsigned int max;		/** Maximal value */
	unsigned int granularity;	/** granularity used in this range */
};

const static struct ontime_range_struct ontime_range[][3] = {
		{
			{10, 500, 10},				// 10 - 500 ns -> 10 ns
			{500, 10000, 100},			// 500 ns - 10 us -> 100 ns
			{0, 0, 0},
		},
		{
			{10, 100, 1},				// 10 us - 100 us -> 1 us
			{100, 1000, 10},			// 100 us - 1 ms -> 10 us
			{1000, 10000, 100}, 		// 1 ms - 10 ms -> 100 us
		},
		{
			{10, 100, 1},				// 10 ms - 100 ms -> 1 ms
			{100, 1000, 10},			// 100 ms - 1 s -> 10 ms
			{1000, 10000, 100},			// 1 s - 10 s -> 100 ms
		},
		{
			{10, 100, 1},				// 10 s - 100 s -> 1 s
			{0, 0, 0},
			{0, 0, 0},
		}
};

/**
 * checks ontime to be in possible range, with useful resolution
 */
void check_ontime(void){
	int i;
	for(i=0;i<3; i++){
		if(ontime_range[timebaserange][i].min <= ontime && ontime < ontime_range[timebaserange][i].max){
			ontime = ontime/ontime_range[timebaserange][i].granularity * ontime_range[timebaserange][i].granularity;
			return;
		}
	}
	if(ontime < ontime_range[timebaserange][0].min) ontime = ontime_range[timebaserange][0].min;
}

/**
 * Checks if framerate is in useful range and switches to CW mode if a duty cycle >= 1 is selected
 */
void check_framerate(void){
	//if(framerate < 0) framerate = 0;
	if(framerate > 50000) framerate = 50000;
	switch(timebaserange){ // case 0: no problem at all
	case 1:
		if(ontime >= 100){
			if(ontime/framerate >= 100)
				framerate = 0;
		}
		break;
	case 2:
		if(ontime*10/framerate)
			framerate = 0;
		break;
	}
}

/**
 * Checks the configured current and limits to a total avg power of 10W
 */
void check_current(void){
	// p_max_hardware ~ 10W
	// limited to 8W to prevent overheating
	const int p_max = 8000; // mW
	const long led_voltage = 20; // ~ 4 * 5
	unsigned long duty;
	unsigned long power;

	if(led_current > 2400) led_current = 2400;
	if(led_current < 200) led_current = 200;

	if(framerate > 0){
		switch(timebaserange){
		case 0: //10 n - 10 u
			duty = (ontime/100)/framerate; // /1000
			power = duty*led_voltage*(led_current/100);
			power /= 10;
			break;
		case 1: // 10 u - 10 m
			duty = ontime/framerate; // /1000
			power = duty*led_voltage*led_current;
			power /= 100;
			break;
		case 2:  // 10 m - 10 s
			if(ontime < 100){
				duty = ontime*1000UL/framerate;
				power = duty*led_voltage*led_current;
				power/= 100;
			}else
				power = led_current * led_voltage;
			break;
		case 3: // don't care about the duty cycle any more, power will most likely not be available
			power = led_current * led_voltage;
			break;
		}
	}else
		power = led_current * led_voltage;

	if(power > p_max && framerate > 0){
		led_current = led_current / (power/p_max);
	}

	if(framerate == 0) {
		while(power > p_max){
			led_current-= 25;
			power = led_current * led_voltage;
		}
	}

	led_current = (led_current / 25) * 25; // round to 25 mA
}
